#include "mat.h"

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
double &Mat::operator()(int i,int j) {
  return *(a+i*n+j);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
const double &Mat::operator()(int i,int j) const {
  return *(a+i*n+j);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
Mat::Mat(int m_,int n_,double *a_=NULL) : m(m_), n(n_) {
  if (a_==NULL) {
    a_is_external = 0;
    a = (double *) malloc(m*n*sizeof(double));
  } else {
    a_is_external = 1;
    a = a_;
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
Mat::~Mat() {
  if (!a_is_external) delete[] a;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void Mat::print(void) const {
  for (int i=0; i<m; i++) {
    for (int j=0; j<n; j++) {
      printf("%f ",(*this)(i,j));
    }
    printf("\n");
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void Mat::matmul(const Mat &a, const Mat &b) {
  assert(a.n == b.m);
  for (int j=0; j < a.m; j++) {
    for (int k=0; k < b.n; k++) {
      double x=0;
      const double *aa=&a(j,0);
      const double *bb=&b(0,k);
      const double *aa_end = aa + a.n;
      while (aa<aa_end) {
	x += (*aa++) * (*bb);
	bb += b.n;
      }
      (*this)(j,k) = x;
    }
  }
}
