cccccccccccccccccccccccccccccccccccccccccccccccc
c      Matrix Multiplication MPI Program       c
c      For this simple version, # of procssors c
c      equals # of columns in matrix           c
c                                              c
c      To run, mpirun -np 4 a.out              c
cccccccccccccccccccccccccccccccccccccccccccccccc

      include 'mpif.h'

      parameter (ncols=100, nrows=100)
      integer a(ncols,nrows), b(ncols,nrows), c(ncols,nrows)
      integer buf(ncols,nrows),ans(ncols,nrows)
      integer myid, root, numprocs, ierr, status(MPI_STATUS_SIZE)
      integer sender, count, tag, ncolproc

      call MPI_INIT(ierr)
      call MPI_COMM_RANK( MPI_COMM_WORLD, myid, ierr )
      call MPI_COMM_SIZE( MPI_COMM_WORLD, numprocs, ierr )

      if(numprocs.ne.4) then
        print *, "Please run this exercise on 4 processors"
        call MPI_FINALIZE(ierr)            
        stop
      endif

      root = 0
      tag = 100
      count = nrows*ncols
      ncolproc = ncols/numprocs
      if (ncolproc*numprocs .ne. ncols) then
         write (*,*) 'Number of cols must be multiple of numprocs'
         stop
      endif

c     Master initializes and then dispatches to others
      if ( myid .eq. root ) then

         do j=1,ncols
            do i=1,nrows
               a(i,j) = 1
               b(i,j) = j
            enddo
         enddo

c        send a to all other process
         call MPI_BCAST(a,count,MPI_INTEGER,root,MPI_COMM_WORLD,ierr)

c        send one column of b to each other process
         jstart = 1
         do j=1,numprocs-1
            jstart = jstart + ncolproc
            call MPI_SEND(b(1,jstart),nrows*ncolproc,MPI_INTEGER,
     *           j,tag,MPI_COMM_WORLD,ierr)
         enddo
         
c        Master does his own part here
         do k=1,ncolproc 
            do i=1,nrows
               ans(i) = 0
               do j=1,ncols
                  ans(i) = ans(i) + a(i,j) * b(i,k)
               enddo
               c(i,k) = ans(i)
            enddo
         enddo

c        then receives answers from others

         jstart=1
         do j=1,numprocs-1
            jstart = jstart + ncolproc
            call MPI_RECV(c(1,jstart), nrows*ncolproc, MPI_INTEGER,
     *           MPI_ANY_SOURCE,MPI_ANY_TAG, MPI_COMM_WORLD,
     *           status, ierr)

            sender = status(MPI_SOURCE)
         enddo
         
         do i=1,nrows
            write(6,*)(c(i,j),j=1,ncols)
         enddo

      ELSE

c        slaves receive a, and one column of b, then compute dot product
         call MPI_BCAST(a,count,MPI_INTEGER,root,MPI_COMM_WORLD,ierr)

         call MPI_RECV(buf, nrows, MPI_INTEGER, root,
     $        MPI_ANY_TAG, MPI_COMM_WORLD, status, ierr)

         do i=1,nrows
            ans(i) = 0
            do j=1,ncols
               ans(i) = ans(i) + a(i,j) * buf(j)
            enddo
         enddo

         call MPI_SEND(ans,nrows,MPI_INTEGER,root,0,MPI_COMM_WORLD,ierr)

      ENDIF

      call MPI_FINALIZE(ierr)

      stop
      end

